@@ -19,7 +19,10 @@ import io.netty.buffer.ByteBuf;
 import io.netty.buffer.ByteBufUtil;
 import io.netty.handler.codec.DateFormatter;
 import io.netty.handler.codec.Headers;
+import io.netty.handler.codec.HeadersUtils;
 import io.netty.util.AsciiString;
+import io.netty.util.CharsetUtil;
+import io.netty.util.internal.ObjectUtil;
 
 import java.text.ParseException;
 import java.util.Calendar;
@@ -30,6 +33,9 @@ import java.util.Map;
 import java.util.Map.Entry;
 import java.util.Set;
 
+import static io.netty.util.AsciiString.contentEquals;
+import static io.netty.util.AsciiString.contentEqualsIgnoreCase;
+import static io.netty.util.AsciiString.trim;
 import static io.netty.util.internal.ObjectUtil.checkNotNull;
 
 /**
@@ -564,7 +570,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #get(CharSequence, String)} instead.
      *
-     * @see {@link #getHeader(HttpMessage, CharSequence, String)}
+     * @see #getHeader(HttpMessage, CharSequence, String)
      */
     @Deprecated
     public static String getHeader(HttpMessage message, String name, String defaultValue) {
@@ -589,7 +595,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #set(CharSequence, Object)} instead.
      *
-     * @see {@link #setHeader(HttpMessage, CharSequence, Object)}
+     * @see #setHeader(HttpMessage, CharSequence, Object)
      */
     @Deprecated
     public static void setHeader(HttpMessage message, String name, Object value) {
@@ -614,7 +620,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #set(CharSequence, Iterable)} instead.
      *
-     * @see {@link #setHeader(HttpMessage, CharSequence, Iterable)}
+     * @see #setHeader(HttpMessage, CharSequence, Iterable)
      */
     @Deprecated
     public static void setHeader(HttpMessage message, String name, Iterable<?> values) {
@@ -645,7 +651,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #add(CharSequence, Object)} instead.
      *
-     * @see {@link #addHeader(HttpMessage, CharSequence, Object)}
+     * @see #addHeader(HttpMessage, CharSequence, Object)
      */
     @Deprecated
     public static void addHeader(HttpMessage message, String name, Object value) {
@@ -669,7 +675,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #remove(CharSequence)} instead.
      *
-     * @see {@link #removeHeader(HttpMessage, CharSequence)}
+     * @see #removeHeader(HttpMessage, CharSequence)
      */
     @Deprecated
     public static void removeHeader(HttpMessage message, String name) {
@@ -699,7 +705,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #getInt(CharSequence)} instead.
      *
-     * @see {@link #getIntHeader(HttpMessage, CharSequence)}
+     * @see #getIntHeader(HttpMessage, CharSequence)
      */
     @Deprecated
     public static int getIntHeader(HttpMessage message, String name) {
@@ -729,7 +735,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #getInt(CharSequence, int)} instead.
      *
-     * @see {@link #getIntHeader(HttpMessage, CharSequence, int)}
+     * @see #getIntHeader(HttpMessage, CharSequence, int)
      */
     @Deprecated
     public static int getIntHeader(HttpMessage message, String name, int defaultValue) {
@@ -754,7 +760,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #setInt(CharSequence, int)} instead.
      *
-     * @see {@link #setIntHeader(HttpMessage, CharSequence, int)}
+     * @see #setIntHeader(HttpMessage, CharSequence, int)
      */
     @Deprecated
     public static void setIntHeader(HttpMessage message, String name, int value) {
@@ -775,7 +781,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #set(CharSequence, Iterable)} instead.
      *
-     * @see {@link #setIntHeader(HttpMessage, CharSequence, Iterable)}
+     * @see #setIntHeader(HttpMessage, CharSequence, Iterable)
      */
     @Deprecated
     public static void setIntHeader(HttpMessage message, String name, Iterable<Integer> values) {
@@ -796,7 +802,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #add(CharSequence, Iterable)} instead.
      *
-     * @see {@link #addIntHeader(HttpMessage, CharSequence, int)}
+     * @see #addIntHeader(HttpMessage, CharSequence, int)
      */
     @Deprecated
     public static void addIntHeader(HttpMessage message, String name, int value) {
@@ -816,7 +822,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #getTimeMillis(CharSequence)} instead.
      *
-     * @see {@link #getDateHeader(HttpMessage, CharSequence)}
+     * @see #getDateHeader(HttpMessage, CharSequence)
      */
     @Deprecated
     public static Date getDateHeader(HttpMessage message, String name) throws ParseException {
@@ -850,7 +856,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #getTimeMillis(CharSequence, long)} instead.
      *
-     * @see {@link #getDateHeader(HttpMessage, CharSequence, Date)}
+     * @see #getDateHeader(HttpMessage, CharSequence, Date)
      */
     @Deprecated
     public static Date getDateHeader(HttpMessage message, String name, Date defaultValue) {
@@ -877,7 +883,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #set(CharSequence, Object)} instead.
      *
-     * @see {@link #setDateHeader(HttpMessage, CharSequence, Date)}
+     * @see #setDateHeader(HttpMessage, CharSequence, Date)
      */
     @Deprecated
     public static void setDateHeader(HttpMessage message, String name, Date value) {
@@ -904,7 +910,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #set(CharSequence, Iterable)} instead.
      *
-     * @see {@link #setDateHeader(HttpMessage, CharSequence, Iterable)}
+     * @see #setDateHeader(HttpMessage, CharSequence, Iterable)
      */
     @Deprecated
     public static void setDateHeader(HttpMessage message, String name, Iterable<Date> values) {
@@ -927,7 +933,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #add(CharSequence, Object)} instead.
      *
-     * @see {@link #addDateHeader(HttpMessage, CharSequence, Date)}
+     * @see #addDateHeader(HttpMessage, CharSequence, Date)
      */
     @Deprecated
     public static void addDateHeader(HttpMessage message, String name, Date value) {
@@ -1014,7 +1020,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     /**
      * @deprecated Use {@link #set(CharSequence, Object)} instead.
      *
-     * @see {@link #setHost(HttpMessage, CharSequence)}
+     * @see #setHost(HttpMessage, CharSequence)
      */
     @Deprecated
     public static void setHost(HttpMessage message, String value) {
@@ -1145,22 +1151,15 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
      */
     @Deprecated
     public static boolean equalsIgnoreCase(CharSequence name1, CharSequence name2) {
-        return AsciiString.contentEqualsIgnoreCase(name1, name2);
-    }
-
-    static void encode(HttpHeaders headers, ByteBuf buf) throws Exception {
-        Iterator<Entry<CharSequence, CharSequence>> iter = headers.iteratorCharSequence();
-        while (iter.hasNext()) {
-            Entry<CharSequence, CharSequence> header = iter.next();
-            HttpHeadersEncoder.encoderHeader(header.getKey(), header.getValue(), buf);
-        }
+        return contentEqualsIgnoreCase(name1, name2);
     }
 
+    @Deprecated
     public static void encodeAscii(CharSequence seq, ByteBuf buf) {
         if (seq instanceof AsciiString) {
             ByteBufUtil.copy((AsciiString) seq, 0, buf, seq.length());
         } else {
-            HttpUtil.encodeAscii0(seq, buf);
+            buf.writeCharSequence(seq, CharsetUtil.US_ASCII);
         }
     }
 
@@ -1298,7 +1297,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     public abstract List<Map.Entry<String, String>> entries();
 
     /**
-     * @see {@link #contains(CharSequence)}
+     * @see #contains(CharSequence)
      */
     public abstract boolean contains(String name);
 
@@ -1315,6 +1314,24 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
      */
     public abstract Iterator<Entry<CharSequence, CharSequence>> iteratorCharSequence();
 
+    /**
+     * Equivalent to {@link #getAll(String)} but it is possible that no intermediate list is generated.
+     * @param name the name of the header to retrieve
+     * @return an {@link Iterator} of header values corresponding to {@code name}.
+     */
+    public Iterator<String> valueStringIterator(CharSequence name) {
+        return getAll(name).iterator();
+    }
+
+    /**
+     * Equivalent to {@link #getAll(String)} but it is possible that no intermediate list is generated.
+     * @param name the name of the header to retrieve
+     * @return an {@link Iterator} of header values corresponding to {@code name}.
+     */
+    public Iterator<? extends CharSequence> valueCharSequenceIterator(CharSequence name) {
+        return valueStringIterator(name);
+    }
+
     /**
      * Checks to see if there is a header with the specified name
      *
@@ -1396,9 +1413,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
      * @return {@code this}
      */
     public HttpHeaders add(HttpHeaders headers) {
-        if (headers == null) {
-            throw new NullPointerException("headers");
-        }
+        ObjectUtil.checkNotNull(headers, "headers");
         for (Map.Entry<String, String> e: headers) {
             add(e.getKey(), e.getValue());
         }
@@ -1552,18 +1567,16 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
      * @see #contains(CharSequence, CharSequence, boolean)
      */
     public boolean contains(String name, String value, boolean ignoreCase) {
-        List<String> values = getAll(name);
-        if (values.isEmpty()) {
-            return false;
-        }
-
-        for (String v: values) {
-            if (ignoreCase) {
-                if (v.equalsIgnoreCase(value)) {
+        Iterator<String> valueIterator = valueStringIterator(name);
+        if (ignoreCase) {
+            while (valueIterator.hasNext()) {
+                if (valueIterator.next().equalsIgnoreCase(value)) {
                     return true;
                 }
-            } else {
-                if (v.equals(value)) {
+            }
+        } else {
+            while (valueIterator.hasNext()) {
+                if (valueIterator.next().equals(value)) {
                     return true;
                 }
             }
@@ -1573,7 +1586,7 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
 
     /**
      * Returns {@code true} if a header with the {@code name} and {@code value} exists, {@code false} otherwise.
-     * This also handles multiple values that are seperated with a {@code ,}.
+     * This also handles multiple values that are separated with a {@code ,}.
      * <p>
      * If {@code ignoreCase} is {@code true} then a case insensitive compare is done on the value.
      * @param name the name of the header to find
@@ -1582,32 +1595,56 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
      * otherwise a case sensitive compare is run to compare values.
      */
     public boolean containsValue(CharSequence name, CharSequence value, boolean ignoreCase) {
-        List<String> values = getAll(name);
-        if (values.isEmpty()) {
-            return false;
-        }
-
-        for (String v: values) {
-            if (contains(v, value, ignoreCase)) {
+        Iterator<? extends CharSequence> itr = valueCharSequenceIterator(name);
+        while (itr.hasNext()) {
+            if (containsCommaSeparatedTrimmed(itr.next(), value, ignoreCase)) {
                 return true;
             }
         }
         return false;
     }
 
-    private static boolean contains(String value, CharSequence expected, boolean ignoreCase) {
-        String[] parts = value.split(",");
+    private static boolean containsCommaSeparatedTrimmed(CharSequence rawNext, CharSequence expected,
+                                                         boolean ignoreCase) {
+        int begin = 0;
+        int end;
         if (ignoreCase) {
-            for (String s: parts) {
-                if (AsciiString.contentEqualsIgnoreCase(expected, s.trim())) {
+            if ((end = AsciiString.indexOf(rawNext, ',', begin)) == -1) {
+                if (contentEqualsIgnoreCase(trim(rawNext), expected)) {
                     return true;
                 }
+            } else {
+                do {
+                    if (contentEqualsIgnoreCase(trim(rawNext.subSequence(begin, end)), expected)) {
+                        return true;
+                    }
+                    begin = end + 1;
+                } while ((end = AsciiString.indexOf(rawNext, ',', begin)) != -1);
+
+                if (begin < rawNext.length()) {
+                    if (contentEqualsIgnoreCase(trim(rawNext.subSequence(begin, rawNext.length())), expected)) {
+                        return true;
+                    }
+                }
             }
         } else {
-            for (String s: parts) {
-                if (AsciiString.contentEquals(expected, s.trim())) {
+            if ((end = AsciiString.indexOf(rawNext, ',', begin)) == -1) {
+                if (contentEquals(trim(rawNext), expected)) {
                     return true;
                 }
+            } else {
+                do {
+                    if (contentEquals(trim(rawNext.subSequence(begin, end)), expected)) {
+                        return true;
+                    }
+                    begin = end + 1;
+                } while ((end = AsciiString.indexOf(rawNext, ',', begin)) != -1);
+
+                if (begin < rawNext.length()) {
+                    if (contentEquals(trim(rawNext.subSequence(begin, rawNext.length())), expected)) {
+                        return true;
+                    }
+                }
             }
         }
         return false;
@@ -1650,4 +1687,16 @@ public abstract class HttpHeaders implements Iterable<Map.Entry<String, String>>
     public boolean contains(CharSequence name, CharSequence value, boolean ignoreCase) {
         return contains(name.toString(), value.toString(), ignoreCase);
     }
+
+    @Override
+    public String toString() {
+        return HeadersUtils.toString(getClass(), iteratorCharSequence(), size());
+    }
+
+    /**
+     * Returns a deep copy of the passed in {@link HttpHeaders}.
+     */
+    public HttpHeaders copy() {
+        return new DefaultHttpHeaders().set(this);
+    }
 }
