@@ -203,6 +203,13 @@ public class ClientCookieDecoderTest {
         assertNull(cookie);
     }
 
+    @Test
+    public void testDecodingInvalidValuesWithCommaAtStart() {
+        assertNull(ClientCookieDecoder.STRICT.decode(","));
+        assertNull(ClientCookieDecoder.STRICT.decode(",a"));
+        assertNull(ClientCookieDecoder.STRICT.decode(",a=a"));
+    }
+
     @Test
     public void testDecodingLongValue() {
         String longValue =
