@@ -22,6 +22,7 @@ import io.netty.channel.ChannelPromise;
 import io.netty.handler.codec.http2.Http2CodecUtil.SimpleChannelPromiseAggregator;
 import io.netty.handler.codec.http2.Http2FrameWriter.Configuration;
 import io.netty.handler.codec.http2.Http2HeadersEncoder.SensitivityDetector;
+import io.netty.util.internal.PlatformDependent;
 import io.netty.util.internal.UnstableApi;
 
 import static io.netty.buffer.Unpooled.directBuffer;
@@ -47,8 +48,6 @@ import static io.netty.handler.codec.http2.Http2CodecUtil.WINDOW_UPDATE_FRAME_LE
 import static io.netty.handler.codec.http2.Http2CodecUtil.isMaxFrameSizeValid;
 import static io.netty.handler.codec.http2.Http2CodecUtil.verifyPadding;
 import static io.netty.handler.codec.http2.Http2CodecUtil.writeFrameHeaderInternal;
-import static io.netty.handler.codec.http2.Http2CodecUtil.writeUnsignedInt;
-import static io.netty.handler.codec.http2.Http2CodecUtil.writeUnsignedShort;
 import static io.netty.handler.codec.http2.Http2Error.FRAME_SIZE_ERROR;
 import static io.netty.handler.codec.http2.Http2Exception.connectionError;
 import static io.netty.handler.codec.http2.Http2FrameTypes.CONTINUATION;
@@ -62,6 +61,8 @@ import static io.netty.handler.codec.http2.Http2FrameTypes.RST_STREAM;
 import static io.netty.handler.codec.http2.Http2FrameTypes.SETTINGS;
 import static io.netty.handler.codec.http2.Http2FrameTypes.WINDOW_UPDATE;
 import static io.netty.util.internal.ObjectUtil.checkNotNull;
+import static io.netty.util.internal.ObjectUtil.checkPositive;
+import static io.netty.util.internal.ObjectUtil.checkPositiveOrZero;
 import static java.lang.Math.max;
 import static java.lang.Math.min;
 
@@ -87,12 +88,12 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
         this(new DefaultHttp2HeadersEncoder());
     }
 
-    public DefaultHttp2FrameWriter(SensitivityDetector headersSensativityDetector) {
-        this(new DefaultHttp2HeadersEncoder(headersSensativityDetector));
+    public DefaultHttp2FrameWriter(SensitivityDetector headersSensitivityDetector) {
+        this(new DefaultHttp2HeadersEncoder(headersSensitivityDetector));
     }
 
-    public DefaultHttp2FrameWriter(SensitivityDetector headersSensativityDetector, boolean ignoreMaxHeaderListSize) {
-        this(new DefaultHttp2HeadersEncoder(headersSensativityDetector, ignoreMaxHeaderListSize));
+    public DefaultHttp2FrameWriter(SensitivityDetector headersSensitivityDetector, boolean ignoreMaxHeaderListSize) {
+        this(new DefaultHttp2HeadersEncoder(headersSensitivityDetector, ignoreMaxHeaderListSize));
     }
 
     public DefaultHttp2FrameWriter(Http2HeadersEncoder headersEncoder) {
@@ -106,8 +107,8 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
     }
 
     @Override
-    public Http2HeaderTable headerTable() {
-        return headersEncoder.configuration().headerTable();
+    public Http2HeadersEncoder.Configuration headersConfiguration() {
+        return headersEncoder.configuration();
     }
 
     @Override
@@ -136,51 +137,119 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
             int padding, boolean endStream, ChannelPromise promise) {
         final SimpleChannelPromiseAggregator promiseAggregator =
                 new SimpleChannelPromiseAggregator(promise, ctx.channel(), ctx.executor());
-        final DataFrameHeader header = new DataFrameHeader(ctx, streamId);
-        boolean needToReleaseHeaders = true;
-        boolean needToReleaseData = true;
+        ByteBuf frameHeader = null;
         try {
             verifyStreamId(streamId, STREAM_ID);
             verifyPadding(padding);
 
-            boolean lastFrame;
             int remainingData = data.readableBytes();
-            do {
-                // Determine how much data and padding to write in this frame. Put all padding at the end.
-                int frameDataBytes = min(remainingData, maxFrameSize);
-                int framePaddingBytes = min(padding, max(0, (maxFrameSize - 1) - frameDataBytes));
+            Http2Flags flags = new Http2Flags();
+            flags.endOfStream(false);
+            flags.paddingPresent(false);
+            // Fast path to write frames of payload size maxFrameSize first.
+            if (remainingData > maxFrameSize) {
+                frameHeader = ctx.alloc().buffer(FRAME_HEADER_LENGTH);
+                writeFrameHeaderInternal(frameHeader, maxFrameSize, DATA, flags, streamId);
+                do {
+                    // Write the header.
+                    ctx.write(frameHeader.retainedSlice(), promiseAggregator.newPromise());
 
-                // Decrement the remaining counters.
-                padding -= framePaddingBytes;
-                remainingData -= frameDataBytes;
+                    // Write the payload.
+                    ctx.write(data.readRetainedSlice(maxFrameSize), promiseAggregator.newPromise());
 
-                // Determine whether or not this is the last frame to be sent.
-                lastFrame = remainingData == 0 && padding == 0;
+                    remainingData -= maxFrameSize;
+                    // Stop iterating if remainingData == maxFrameSize so we can take care of reference counts below.
+                } while (remainingData > maxFrameSize);
+            }
 
-                // Only the last frame is not retained. Until then, the outer finally must release.
-                ByteBuf frameHeader = header.slice(frameDataBytes, framePaddingBytes, lastFrame && endStream);
-                needToReleaseHeaders = !lastFrame;
-                ctx.write(lastFrame ? frameHeader : frameHeader.retain(), promiseAggregator.newPromise());
-
-                // Write the frame data.
-                ByteBuf frameData = data.readSlice(frameDataBytes);
-                // Only the last frame is not retained. Until then, the outer finally must release.
-                needToReleaseData = !lastFrame;
-                ctx.write(lastFrame ? frameData : frameData.retain(), promiseAggregator.newPromise());
-
-                // Write the frame padding.
-                if (paddingBytes(framePaddingBytes) > 0) {
-                    ctx.write(ZERO_BUFFER.slice(0, paddingBytes(framePaddingBytes)), promiseAggregator.newPromise());
+            if (padding == 0) {
+                // Write the header.
+                if (frameHeader != null) {
+                    frameHeader.release();
+                    frameHeader = null;
                 }
-            } while (!lastFrame);
-        } catch (Throwable t) {
-            if (needToReleaseHeaders) {
-                header.release();
+                ByteBuf frameHeader2 = ctx.alloc().buffer(FRAME_HEADER_LENGTH);
+                flags.endOfStream(endStream);
+                writeFrameHeaderInternal(frameHeader2, remainingData, DATA, flags, streamId);
+                ctx.write(frameHeader2, promiseAggregator.newPromise());
+
+                // Write the payload.
+                ByteBuf lastFrame = data.readSlice(remainingData);
+                data = null;
+                ctx.write(lastFrame, promiseAggregator.newPromise());
+            } else {
+                if (remainingData != maxFrameSize) {
+                    if (frameHeader != null) {
+                        frameHeader.release();
+                        frameHeader = null;
+                    }
+                } else {
+                    remainingData -= maxFrameSize;
+                    // Write the header.
+                    ByteBuf lastFrame;
+                    if (frameHeader == null) {
+                        lastFrame = ctx.alloc().buffer(FRAME_HEADER_LENGTH);
+                        writeFrameHeaderInternal(lastFrame, maxFrameSize, DATA, flags, streamId);
+                    } else {
+                        lastFrame = frameHeader.slice();
+                        frameHeader = null;
+                    }
+                    ctx.write(lastFrame, promiseAggregator.newPromise());
+
+                    // Write the payload.
+                    lastFrame = data.readableBytes() != maxFrameSize ? data.readSlice(maxFrameSize) : data;
+                    data = null;
+                    ctx.write(lastFrame, promiseAggregator.newPromise());
+                }
+
+                do {
+                    int frameDataBytes = min(remainingData, maxFrameSize);
+                    int framePaddingBytes = min(padding, max(0, (maxFrameSize - 1) - frameDataBytes));
+
+                    // Decrement the remaining counters.
+                    padding -= framePaddingBytes;
+                    remainingData -= frameDataBytes;
+
+                    // Write the header.
+                    ByteBuf frameHeader2 = ctx.alloc().buffer(DATA_FRAME_HEADER_LENGTH);
+                    flags.endOfStream(endStream && remainingData == 0 && padding == 0);
+                    flags.paddingPresent(framePaddingBytes > 0);
+                    writeFrameHeaderInternal(frameHeader2, framePaddingBytes + frameDataBytes, DATA, flags, streamId);
+                    writePaddingLength(frameHeader2, framePaddingBytes);
+                    ctx.write(frameHeader2, promiseAggregator.newPromise());
+
+                    // Write the payload.
+                    if (frameDataBytes != 0) {
+                        if (remainingData == 0) {
+                            ByteBuf lastFrame = data.readSlice(frameDataBytes);
+                            data = null;
+                            ctx.write(lastFrame, promiseAggregator.newPromise());
+                        } else {
+                            ctx.write(data.readRetainedSlice(frameDataBytes), promiseAggregator.newPromise());
+                        }
+                    }
+                    // Write the frame padding.
+                    if (paddingBytes(framePaddingBytes) > 0) {
+                        ctx.write(ZERO_BUFFER.slice(0, paddingBytes(framePaddingBytes)),
+                                  promiseAggregator.newPromise());
+                    }
+                } while (remainingData != 0 || padding != 0);
             }
-            if (needToReleaseData) {
-                data.release();
+        } catch (Throwable cause) {
+            if (frameHeader != null) {
+                frameHeader.release();
             }
-            promiseAggregator.setFailure(t);
+            // Use a try/finally here in case the data has been released before calling this method. This is not
+            // necessary above because we internally allocate frameHeader.
+            try {
+                if (data != null) {
+                    data.release();
+                }
+            } finally {
+                promiseAggregator.setFailure(cause);
+                promiseAggregator.doneAllocatingPromises();
+            }
+            return promiseAggregator;
         }
         return promiseAggregator.doneAllocatingPromises();
     }
@@ -210,8 +279,7 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
 
             ByteBuf buf = ctx.alloc().buffer(PRIORITY_FRAME_LENGTH);
             writeFrameHeaderInternal(buf, PRIORITY_ENTRY_LENGTH, PRIORITY, new Http2Flags(), streamId);
-            long word1 = exclusive ? 0x80000000L | streamDependency : streamDependency;
-            writeUnsignedInt(word1, buf);
+            buf.writeInt(exclusive ? (int) (0x80000000L | streamDependency) : streamDependency);
             // Adjust the weight so that it fits into a single byte on the wire.
             buf.writeByte(weight - 1);
             return ctx.write(buf, promise);
@@ -229,7 +297,7 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
 
             ByteBuf buf = ctx.alloc().buffer(RST_STREAM_FRAME_LENGTH);
             writeFrameHeaderInternal(buf, INT_FIELD_LENGTH, RST_STREAM, new Http2Flags(), streamId);
-            writeUnsignedInt(errorCode, buf);
+            buf.writeInt((int) errorCode);
             return ctx.write(buf, promise);
         } catch (Throwable t) {
             return promise.setFailure(t);
@@ -245,8 +313,8 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
             ByteBuf buf = ctx.alloc().buffer(FRAME_HEADER_LENGTH + settings.size() * SETTING_ENTRY_LENGTH);
             writeFrameHeaderInternal(buf, payloadLength, SETTINGS, new Http2Flags(), 0);
             for (Http2Settings.PrimitiveEntry<Long> entry : settings.entries()) {
-                writeUnsignedShort(entry.key(), buf);
-                writeUnsignedInt(entry.value(), buf);
+                buf.writeChar(entry.key());
+                buf.writeInt(entry.value().intValue());
             }
             return ctx.write(buf, promise);
         } catch (Throwable t) {
@@ -266,28 +334,14 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
     }
 
     @Override
-    public ChannelFuture writePing(ChannelHandlerContext ctx, boolean ack, ByteBuf data,
-            ChannelPromise promise) {
-        boolean releaseData = true;
-        SimpleChannelPromiseAggregator promiseAggregator =
-                new SimpleChannelPromiseAggregator(promise, ctx.channel(), ctx.executor());
-        try {
-            verifyPingPayload(data);
-            Http2Flags flags = ack ? new Http2Flags().ack(true) : new Http2Flags();
-            ByteBuf buf = ctx.alloc().buffer(FRAME_HEADER_LENGTH);
-            writeFrameHeaderInternal(buf, data.readableBytes(), PING, flags, 0);
-            ctx.write(buf, promiseAggregator.newPromise());
-
-            // Write the debug data.
-            releaseData = false;
-            ctx.write(data, promiseAggregator.newPromise());
-        } catch (Throwable t) {
-            if (releaseData) {
-                data.release();
-            }
-            promiseAggregator.setFailure(t);
-        }
-        return promiseAggregator.doneAllocatingPromises();
+    public ChannelFuture writePing(ChannelHandlerContext ctx, boolean ack, long data, ChannelPromise promise) {
+        Http2Flags flags = ack ? new Http2Flags().ack(true) : new Http2Flags();
+        ByteBuf buf = ctx.alloc().buffer(FRAME_HEADER_LENGTH + PING_FRAME_PAYLOAD_LENGTH);
+        // Assume nothing below will throw until buf is written. That way we don't have to take care of ownership
+        // in the catch block.
+        writeFrameHeaderInternal(buf, PING_FRAME_PAYLOAD_LENGTH, PING, flags, 0);
+        buf.writeLong(data);
+        return ctx.write(buf, promise);
     }
 
     @Override
@@ -332,10 +386,14 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
             }
 
             if (!flags.endOfHeaders()) {
-                writeContinuationFrames(ctx, streamId, headerBlock, padding, promiseAggregator);
+                writeContinuationFrames(ctx, streamId, headerBlock, promiseAggregator);
             }
+        } catch (Http2Exception e) {
+            promiseAggregator.setFailure(e);
         } catch (Throwable t) {
             promiseAggregator.setFailure(t);
+            promiseAggregator.doneAllocatingPromises();
+            PlatformDependent.throwException(t);
         } finally {
             if (headerBlock != null) {
                 headerBlock.release();
@@ -347,7 +405,6 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
     @Override
     public ChannelFuture writeGoAway(ChannelHandlerContext ctx, int lastStreamId, long errorCode,
             ByteBuf debugData, ChannelPromise promise) {
-        boolean releaseData = true;
         SimpleChannelPromiseAggregator promiseAggregator =
                 new SimpleChannelPromiseAggregator(promise, ctx.channel(), ctx.executor());
         try {
@@ -356,17 +413,25 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
 
             int payloadLength = 8 + debugData.readableBytes();
             ByteBuf buf = ctx.alloc().buffer(GO_AWAY_FRAME_HEADER_LENGTH);
+            // Assume nothing below will throw until buf is written. That way we don't have to take care of ownership
+            // in the catch block.
             writeFrameHeaderInternal(buf, payloadLength, GO_AWAY, new Http2Flags(), 0);
             buf.writeInt(lastStreamId);
-            writeUnsignedInt(errorCode, buf);
+            buf.writeInt((int) errorCode);
             ctx.write(buf, promiseAggregator.newPromise());
+        } catch (Throwable t) {
+            try {
+                debugData.release();
+            } finally {
+                promiseAggregator.setFailure(t);
+                promiseAggregator.doneAllocatingPromises();
+            }
+            return promiseAggregator;
+        }
 
-            releaseData = false;
+        try {
             ctx.write(debugData, promiseAggregator.newPromise());
         } catch (Throwable t) {
-            if (releaseData) {
-                debugData.release();
-            }
             promiseAggregator.setFailure(t);
         }
         return promiseAggregator.doneAllocatingPromises();
@@ -391,21 +456,27 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
     @Override
     public ChannelFuture writeFrame(ChannelHandlerContext ctx, byte frameType, int streamId,
             Http2Flags flags, ByteBuf payload, ChannelPromise promise) {
-        boolean releaseData = true;
         SimpleChannelPromiseAggregator promiseAggregator =
                 new SimpleChannelPromiseAggregator(promise, ctx.channel(), ctx.executor());
         try {
             verifyStreamOrConnectionId(streamId, STREAM_ID);
             ByteBuf buf = ctx.alloc().buffer(FRAME_HEADER_LENGTH);
+            // Assume nothing below will throw until buf is written. That way we don't have to take care of ownership
+            // in the catch block.
             writeFrameHeaderInternal(buf, payload.readableBytes(), frameType, flags, streamId);
             ctx.write(buf, promiseAggregator.newPromise());
-
-            releaseData = false;
+        } catch (Throwable t) {
+            try {
+                payload.release();
+            } finally {
+                promiseAggregator.setFailure(t);
+                promiseAggregator.doneAllocatingPromises();
+            }
+            return promiseAggregator;
+        }
+        try {
             ctx.write(payload, promiseAggregator.newPromise());
         } catch (Throwable t) {
-            if (releaseData) {
-                payload.release();
-            }
             promiseAggregator.setFailure(t);
         }
         return promiseAggregator.doneAllocatingPromises();
@@ -446,8 +517,7 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
             writePaddingLength(buf, padding);
 
             if (hasPriority) {
-                long word1 = exclusive ? 0x80000000L | streamDependency : streamDependency;
-                writeUnsignedInt(word1, buf);
+                buf.writeInt(exclusive ? (int) (0x80000000L | streamDependency) : streamDependency);
 
                 // Adjust the weight so that it fits into a single byte on the wire.
                 buf.writeByte(weight - 1);
@@ -463,10 +533,14 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
             }
 
             if (!flags.endOfHeaders()) {
-                writeContinuationFrames(ctx, streamId, headerBlock, padding, promiseAggregator);
+                writeContinuationFrames(ctx, streamId, headerBlock, promiseAggregator);
             }
+        } catch (Http2Exception e) {
+            promiseAggregator.setFailure(e);
         } catch (Throwable t) {
             promiseAggregator.setFailure(t);
+            promiseAggregator.doneAllocatingPromises();
+            PlatformDependent.throwException(t);
         } finally {
             if (headerBlock != null) {
                 headerBlock.release();
@@ -479,28 +553,19 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
      * Writes as many continuation frames as needed until {@code padding} and {@code headerBlock} are consumed.
      */
     private ChannelFuture writeContinuationFrames(ChannelHandlerContext ctx, int streamId,
-            ByteBuf headerBlock, int padding, SimpleChannelPromiseAggregator promiseAggregator) {
-        Http2Flags flags = new Http2Flags().paddingPresent(padding > 0);
-        int maxFragmentLength = maxFrameSize - padding;
-        // TODO: same padding is applied to all frames, is this desired?
-        if (maxFragmentLength <= 0) {
-            return promiseAggregator.setFailure(new IllegalArgumentException(
-                    "Padding [" + padding + "] is too large for max frame size [" + maxFrameSize + "]"));
-        }
+            ByteBuf headerBlock, SimpleChannelPromiseAggregator promiseAggregator) {
+        Http2Flags flags = new Http2Flags();
 
         if (headerBlock.isReadable()) {
             // The frame header (and padding) only changes on the last frame, so allocate it once and re-use
-            int fragmentReadableBytes = min(headerBlock.readableBytes(), maxFragmentLength);
-            int payloadLength = fragmentReadableBytes + padding;
+            int fragmentReadableBytes = min(headerBlock.readableBytes(), maxFrameSize);
             ByteBuf buf = ctx.alloc().buffer(CONTINUATION_FRAME_HEADER_LENGTH);
-            writeFrameHeaderInternal(buf, payloadLength, CONTINUATION, flags, streamId);
-            writePaddingLength(buf, padding);
+            writeFrameHeaderInternal(buf, fragmentReadableBytes, CONTINUATION, flags, streamId);
 
             do {
-                fragmentReadableBytes = min(headerBlock.readableBytes(), maxFragmentLength);
+                fragmentReadableBytes = min(headerBlock.readableBytes(), maxFrameSize);
                 ByteBuf fragment = headerBlock.readRetainedSlice(fragmentReadableBytes);
 
-                payloadLength = fragmentReadableBytes + padding;
                 if (headerBlock.isReadable()) {
                     ctx.write(buf.retain(), promiseAggregator.newPromise());
                 } else {
@@ -508,18 +573,13 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
                     flags = flags.endOfHeaders(true);
                     buf.release();
                     buf = ctx.alloc().buffer(CONTINUATION_FRAME_HEADER_LENGTH);
-                    writeFrameHeaderInternal(buf, payloadLength, CONTINUATION, flags, streamId);
-                    writePaddingLength(buf, padding);
+                    writeFrameHeaderInternal(buf, fragmentReadableBytes, CONTINUATION, flags, streamId);
                     ctx.write(buf, promiseAggregator.newPromise());
                 }
 
                 ctx.write(fragment, promiseAggregator.newPromise());
 
-                // Write out the padding, if any.
-                if (paddingBytes(padding) > 0) {
-                    ctx.write(ZERO_BUFFER.slice(0, paddingBytes(padding)), promiseAggregator.newPromise());
-                }
-            } while(headerBlock.isReadable());
+            } while (headerBlock.isReadable());
         }
         return promiseAggregator;
     }
@@ -542,15 +602,11 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
     }
 
     private static void verifyStreamId(int streamId, String argumentName) {
-        if (streamId <= 0) {
-            throw new IllegalArgumentException(argumentName + " must be > 0");
-        }
+        checkPositive(streamId, "streamId");
     }
 
     private static void verifyStreamOrConnectionId(int streamId, String argumentName) {
-        if (streamId < 0) {
-            throw new IllegalArgumentException(argumentName + " must be >= 0");
-        }
+        checkPositiveOrZero(streamId, "streamId");
     }
 
     private static void verifyWeight(short weight) {
@@ -566,9 +622,7 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
     }
 
     private static void verifyWindowSizeIncrement(int windowSizeIncrement) {
-        if (windowSizeIncrement < 0) {
-            throw new IllegalArgumentException("WindowSizeIncrement must be >= 0");
-        }
+        checkPositiveOrZero(windowSizeIncrement, "windowSizeIncrement");
     }
 
     private static void verifyPingPayload(ByteBuf data) {
@@ -576,51 +630,4 @@ public class DefaultHttp2FrameWriter implements Http2FrameWriter, Http2FrameSize
             throw new IllegalArgumentException("Opaque data must be " + PING_FRAME_PAYLOAD_LENGTH + " bytes");
         }
     }
-
-    /**
-     * Utility class that manages the creation of frame header buffers for {@code DATA} frames. Attempts
-     * to reuse the same buffer repeatedly when splitting data into multiple frames.
-     */
-    private static final class DataFrameHeader {
-        private final int streamId;
-        private final ByteBuf buffer;
-        private final Http2Flags flags = new Http2Flags();
-        private int prevData;
-        private int prevPadding;
-        private ByteBuf frameHeader;
-
-        DataFrameHeader(ChannelHandlerContext ctx, int streamId) {
-            // All padding will be put at the end, so in the worst case we need 3 headers:
-            // a repeated no-padding frame of maxFrameSize, a frame that has part data and part
-            // padding, and a frame that has the remainder of the padding.
-            buffer = ctx.alloc().buffer(3 * DATA_FRAME_HEADER_LENGTH);
-            this.streamId = streamId;
-        }
-
-        /**
-         * Gets the frame header buffer configured for the current frame.
-         */
-        ByteBuf slice(int data, int padding, boolean endOfStream) {
-            // Since we're reusing the current frame header whenever possible, check if anything changed
-            // that requires a new header.
-            if (data != prevData || padding != prevPadding
-                    || endOfStream != flags.endOfStream() || frameHeader == null) {
-                // Update the header state.
-                prevData = data;
-                prevPadding = padding;
-                flags.paddingPresent(padding > 0);
-                flags.endOfStream(endOfStream);
-                frameHeader = buffer.readSlice(DATA_FRAME_HEADER_LENGTH).writerIndex(0);
-
-                int payloadLength = data + padding;
-                writeFrameHeaderInternal(frameHeader, payloadLength, DATA, flags, streamId);
-                writePaddingLength(frameHeader, padding);
-            }
-            return frameHeader.slice();
-        }
-
-        void release() {
-            buffer.release();
-        }
-    }
 }
