@@ -40,7 +40,6 @@ import java.util.LinkedList;
 import java.util.List;
 
 import static io.netty.buffer.Unpooled.EMPTY_BUFFER;
-import static io.netty.handler.codec.http2.Http2CodecUtil.DEFAULT_HEADER_LIST_SIZE;
 import static io.netty.handler.codec.http2.Http2CodecUtil.MAX_PADDING;
 import static io.netty.handler.codec.http2.Http2HeadersEncoder.NEVER_SENSITIVE;
 import static io.netty.handler.codec.http2.Http2TestUtil.newTestDecoder;
@@ -51,12 +50,12 @@ import static java.lang.Math.min;
 import static org.junit.Assert.assertEquals;
 import static org.junit.Assert.assertFalse;
 import static org.junit.Assert.fail;
-import static org.mockito.Matchers.any;
-import static org.mockito.Matchers.anyBoolean;
-import static org.mockito.Matchers.anyInt;
-import static org.mockito.Matchers.anyShort;
-import static org.mockito.Matchers.eq;
-import static org.mockito.Matchers.isA;
+import static org.mockito.Mockito.any;
+import static org.mockito.Mockito.anyBoolean;
+import static org.mockito.Mockito.anyInt;
+import static org.mockito.Mockito.anyShort;
+import static org.mockito.Mockito.eq;
+import static org.mockito.Mockito.isA;
 import static org.mockito.Mockito.atLeastOnce;
 import static org.mockito.Mockito.doAnswer;
 import static org.mockito.Mockito.never;
@@ -89,7 +88,7 @@ public class Http2FrameRoundtripTest {
 
     private Http2FrameWriter writer;
     private Http2FrameReader reader;
-    private List<ByteBuf> needReleasing = new LinkedList<ByteBuf>();
+    private final List<ByteBuf> needReleasing = new LinkedList<ByteBuf>();
 
     @Before
     public void setup() throws Exception {
@@ -283,7 +282,9 @@ public class Http2FrameRoundtripTest {
     @Test
     public void headersThatAreTooBigShouldFail() throws Exception {
         reader = new DefaultHttp2FrameReader(false);
-        final Http2Headers headers = headersOfSize(DEFAULT_HEADER_LIST_SIZE + 1);
+        final int maxListSize = 100;
+        reader.configuration().headersConfiguration().maxHeaderListSize(maxListSize, maxListSize);
+        final Http2Headers headers = headersOfSize(maxListSize + 1);
         writer.writeHeaders(ctx, STREAM_ID, headers, 2, (short) 3, true, MAX_PADDING, true, ctx.newPromise());
         try {
             readFrames();
@@ -350,26 +351,22 @@ public class Http2FrameRoundtripTest {
 
     @Test
     public void pingFrameShouldMatch() throws Exception {
-        final ByteBuf data = buf("01234567".getBytes(UTF_8));
-
-        writer.writePing(ctx, false, data.slice(), ctx.newPromise());
+        writer.writePing(ctx, false, 1234567, ctx.newPromise());
         readFrames();
 
-        ArgumentCaptor<ByteBuf> captor = ArgumentCaptor.forClass(ByteBuf.class);
+        ArgumentCaptor<Long> captor = ArgumentCaptor.forClass(long.class);
         verify(listener).onPingRead(eq(ctx), captor.capture());
-        assertEquals(data, captor.getValue());
+        assertEquals(1234567, (long) captor.getValue());
     }
 
     @Test
     public void pingAckFrameShouldMatch() throws Exception {
-        final ByteBuf data = buf("01234567".getBytes(UTF_8));
-
-        writer.writePing(ctx, true, data.slice(), ctx.newPromise());
+        writer.writePing(ctx, true, 1234567, ctx.newPromise());
         readFrames();
 
-        ArgumentCaptor<ByteBuf> captor = ArgumentCaptor.forClass(ByteBuf.class);
+        ArgumentCaptor<Long> captor = ArgumentCaptor.forClass(long.class);
         verify(listener).onPingAckRead(eq(ctx), captor.capture());
-        assertEquals(data, captor.getValue());
+        assertEquals(1234567, (long) captor.getValue());
     }
 
     @Test
@@ -427,7 +424,7 @@ public class Http2FrameRoundtripTest {
         reader.readFrame(ctx, write, listener);
     }
 
-    private ByteBuf data(int size) {
+    private static ByteBuf data(int size) {
         byte[] data = new byte[size];
         for (int ix = 0; ix < data.length;) {
             int length = min(MESSAGE.length, data.length - ix);
@@ -437,7 +434,7 @@ public class Http2FrameRoundtripTest {
         return buf(data);
     }
 
-    private ByteBuf buf(byte[] bytes) {
+    private static ByteBuf buf(byte[] bytes) {
         return Unpooled.wrappedBuffer(bytes);
     }
 
@@ -473,7 +470,7 @@ public class Http2FrameRoundtripTest {
         return headers;
     }
 
-    private Http2Headers headersOfSize(final int minSize) {
+    private static Http2Headers headersOfSize(final int minSize) {
         final AsciiString singleByte = new AsciiString(new byte[]{0}, false);
         DefaultHttp2Headers headers = new DefaultHttp2Headers(false);
         for (int size = 0; size < minSize; size += 2) {
