@@ -15,6 +15,8 @@
  */
 package io.netty.handler.codec;
 
+import static io.netty.util.internal.ObjectUtil.checkPositiveOrZero;
+
 import io.netty.buffer.ByteBuf;
 import io.netty.channel.ChannelHandler.Sharable;
 import io.netty.channel.ChannelHandlerContext;
@@ -148,9 +150,7 @@ public class LengthFieldPrepender extends MessageToMessageEncoder<ByteBuf> {
                     "lengthFieldLength must be either 1, 2, 3, 4, or 8: " +
                     lengthFieldLength);
         }
-        ObjectUtil.checkNotNull(byteOrder, "byteOrder");
-
-        this.byteOrder = byteOrder;
+        this.byteOrder = ObjectUtil.checkNotNull(byteOrder, "byteOrder");
         this.lengthFieldLength = lengthFieldLength;
         this.lengthIncludesLengthFieldLength = lengthIncludesLengthFieldLength;
         this.lengthAdjustment = lengthAdjustment;
@@ -163,10 +163,7 @@ public class LengthFieldPrepender extends MessageToMessageEncoder<ByteBuf> {
             length += lengthFieldLength;
         }
 
-        if (length < 0) {
-            throw new IllegalArgumentException(
-                    "Adjusted frame length (" + length + ") is less than zero");
-        }
+        checkPositiveOrZero(length, "length");
 
         switch (lengthFieldLength) {
         case 1:
