@@ -15,6 +15,8 @@
  */
 package io.netty.handler.timeout;
 
+import io.netty.util.internal.PlatformDependent;
+
 /**
  * A {@link TimeoutException} raised by {@link ReadTimeoutHandler} when no data
  * was read within a certain period of time.
@@ -23,7 +25,12 @@ public final class ReadTimeoutException extends TimeoutException {
 
     private static final long serialVersionUID = 169287984113283421L;
 
-    public static final ReadTimeoutException INSTANCE = new ReadTimeoutException();
+    public static final ReadTimeoutException INSTANCE = PlatformDependent.javaVersion() >= 7 ?
+            new ReadTimeoutException(true) : new ReadTimeoutException();
 
-    private ReadTimeoutException() { }
+    ReadTimeoutException() { }
+
+    private ReadTimeoutException(boolean shared) {
+        super(shared);
+    }
 }
