@@ -15,11 +15,14 @@
  */
 package io.netty.resolver;
 
+import io.netty.util.internal.UnstableApi;
+
 import java.net.InetAddress;
 
 /**
  * Resolves a hostname against the hosts file entries.
  */
+@UnstableApi
 public interface HostsFileEntriesResolver {
 
     /**
@@ -27,5 +30,11 @@ public interface HostsFileEntriesResolver {
      */
     HostsFileEntriesResolver DEFAULT = new DefaultHostsFileEntriesResolver();
 
-    InetAddress address(String inetHost);
+    /**
+     * Resolve the address of a hostname against the entries in a hosts file, depending on some address types.
+     * @param inetHost the hostname to resolve
+     * @param resolvedAddressTypes the address types to resolve
+     * @return the first matching address
+     */
+    InetAddress address(String inetHost, ResolvedAddressTypes resolvedAddressTypes);
 }
