@@ -16,6 +16,7 @@
 package io.netty.testsuite.util;
 
 import io.netty.util.CharsetUtil;
+import io.netty.util.internal.ObjectUtil;
 import io.netty.util.internal.logging.InternalLogger;
 import io.netty.util.internal.logging.InternalLoggerFactory;
 import org.junit.rules.TestName;
@@ -112,9 +113,7 @@ public final class TestUtils {
 
     public static void dump(String filenamePrefix) throws IOException {
 
-        if (filenamePrefix == null) {
-            throw new NullPointerException("filenamePrefix");
-        }
+        ObjectUtil.checkNotNull(filenamePrefix, "filenamePrefix");
 
         final String timestamp = timestamp();
         final File heapDumpFile = new File(filenamePrefix + '.' + timestamp + ".hprof");
